import re

import pytest_archon
from pytest_archon import archrule
from pytest_archon.failure import pop_failures
from pytest_archon.plugin import format_failures


def test_rule_basic():
    (archrule("basic rule").match("*.collect").should_not_import("pytest_archon.rule").check(pytest_archon))


def test_rule_exclusion():
    (
        archrule("rule exclusion")
        .exclude("pytest_archon")
        .match("*")
        .exclude("pytest_archon.plugin")
        .should_not_import("pytest_archon.rule")
        .check("pytest_archon")
    )


def test_rule_exclusion_regex():
    (
        archrule("rule exclusion", use_regex=True)
        .exclude(r"^pytest_archon$")
        .match(".*")
        .exclude(r"^pytest_archon\..lugin")
        .should_not_import(r"^pytest\warchon\.rule$")
        .check("pytest_archon")
    )


def test_rule_should_import():
    (
        archrule("rule exclusion")
        .match("pytest_archon.plugin")
        .should_import("pytest_archon.rule")
        .check(pytest_archon)
    )


def test_rule_should_check_predicate1():
    def have_more_than_2_deps(m, di, ai):
        return len(di) > 2

    (
        archrule("rule exclusion")
        .match("pytest_archon.rule")
        .should(have_more_than_2_deps)
        .check(pytest_archon)
    )


def test_rule_should_check_predicate2():
    def have_collect_import(m, di, ai):
        return "pytest_archon.collect" in di

    (archrule("rule exclusion").match("pytest_archon.rule").should(have_collect_import).check(pytest_archon))


def test_failing_predicate(create_testset):
    def not_include_module_b(m, di, ai):
        return "abcz.moduleB" not in di

    create_testset(
        ("abcz/__init__.py", ""),
        ("abcz/moduleA.py", "import abcz.moduleB"),
    )
    (archrule("rule constraint").match("abcz.moduleA").should(not_include_module_b).check("abcz"))

    failures = pop_failures()
    longrepr = format_failures(failures)

    assert failures

    assert "FAILED Rule 'rule constraint':" in longrepr
    assert "- module 'abcz.moduleA' VIOLATED constraint 'not_include_module_b'" in longrepr


def test_rule_should_import_list():
    (
        archrule("rule exclusion")
        .match("pytest_archon.plugin")
        .should_import("pytest_archon.rule", "pytest")
        .check(pytest_archon)
    )


def test_toplevel_imports_only():
    (
        archrule("rule exclusion")
        .match("pytest_archon.plugin")
        .should_import("pytest_archon.rule")
        .check(pytest_archon, only_toplevel_imports=True)
    )


def test_only_direct():
    (
        archrule("rule exclusion")
        .match("pytest_archon.plugin")
        .should_not_import("pytest_archon.collect")
        .check("pytest_archon", only_direct_imports=True)
    )


def test_transitive_dependency_succeeds(create_testset):
    create_testset(
        ("abcz/__init__.py", ""),
        ("abcz/moduleA.py", "import abcz.moduleB"),
        ("abcz/moduleB.py", "import abcz.moduleC"),
        ("abcz/moduleC.py", "import abcz.moduleD"),
        ("abcz/moduleD.py", ""),
    )
    archrule("rule exclusion").match("abcz.moduleA").should_import("abcz.moduleD").check("abcz")


def test_transitive_dependency_via_may_import_succeeds(create_testset):
    # may_import does not protect against forbidden imports from transitive dependencies
    create_testset(
        ("abcz/__init__.py", ""),
        ("abcz/moduleA.py", "import abcz.moduleB"),
        ("abcz/moduleB.py", "import abcz.moduleC"),
        ("abcz/moduleC.py", "import abcz.moduleD"),
        ("abcz/moduleD.py", ""),
    )
    (
        archrule("rule exclusion")
        .match("abcz.moduleA")
        .may_import("abcz.moduleC")
        .should_not_import("abcz.moduleD")
        .check("abcz")
    )
    failures = pop_failures()
    assert len(failures) == 1
    assert re.search(r"abcz\.moduleA.*has FORBIDDEN import abcz\.moduleD", failures[0].reason)


def test_bug_may_import_allows_other_modules_pt1(create_testset):
    create_testset(
        ("abcz/__init__.py", ""),
        ("abcz/common/util.py", "import abcz.app\nimport abcz.common.date"),
        ("abcz/common/date.py", ""),
        ("abcz/app.py", ""),
    )

    (
        archrule("common has no dependencies")
        .match("abcz.common*")
        .should_not_import("abcz*")
        .may_import("abcz.common*")
        .check("abcz")
    )
    failures = pop_failures()
    assert len(failures) == 1
    assert re.search(r"abcz\.common\.util.*has FORBIDDEN import abcz\.app", failures[0].reason)


def test_bug_may_import_allows_other_modules_pt2(create_testset):
    create_testset(
        ("abcz/__init__.py", ""),
        ("abcz/common/util.py", "import abcz.app"),
        ("abcz/app.py", ""),
    )

    (archrule("common has no dependencies").match("abcz.common*").should_not_import("abcz*").check("abcz"))
    failures = pop_failures()
    assert len(failures) == 1
    assert re.search(r"abcz\.common\.util.*has FORBIDDEN import abcz\.app", failures[0].reason)


def test_bug_may_import_allows_other_modules_pt3(create_testset):
    create_testset(
        ("abcz/__init__.py", ""),
        ("abcz/common/util.py", "import abcz.app\nimport abcz.common.date"),
        ("abcz/common/date.py", ""),
        ("abcz/app.py", ""),
    )

    (archrule("common has no dependencies").match("abcz.common*").should_not_import("abcz*").check("abcz"))
    failures = pop_failures()
    assert len(failures) == 2
    assert any(
        re.search(r"abcz\.common\.util.*has FORBIDDEN import abcz\.app", failure.reason)
        for failure in failures
    )
    assert any(
        re.search(r"abcz\.common\.util.*has FORBIDDEN import abcz\.common\.date", failure.reason)
        for failure in failures
    )


def test_required_transitive_dependency_fails(create_testset):
    create_testset(
        ("abcz/__init__.py", ""),
        ("abcz/moduleA.py", "import abcz.moduleB"),
        ("abcz/moduleB.py", "import abcz.moduleC"),
        ("abcz/moduleC.py", ""),
    )
    (archrule("rule exclusion").match("abcz.moduleA").should_import("abcz.moduleD").check("abcz"))

    failures = pop_failures()
    longrepr = format_failures(failures)

    assert failures

    assert "FAILED Rule 'rule exclusion':" in longrepr
    assert (
        "- module 'abcz.moduleA' is missing REQUIRED imports matching glob pattern /abcz.moduleD/" in longrepr
    )


def test_forbidden_transitive_dependency_fails(create_testset):
    create_testset(
        ("abcz/__init__.py", ""),
        ("abcz/moduleA.py", "import abcz.moduleB"),
        ("abcz/moduleB.py", "import abcz.moduleC"),
        ("abcz/moduleC.py", "import abcz.moduleD"),
        ("abcz/moduleD.py", ""),
    )
    (archrule("rule exclusion").match("abcz.moduleA").should_not_import("abcz.moduleD").check("abcz"))

    failures = pop_failures()
    longrepr = format_failures(failures)

    assert failures
    assert longrepr
    assert "FAILED Rule 'rule exclusion':" in longrepr
    assert "- module 'abcz.moduleA' has FORBIDDEN import" in longrepr
    assert "abcz.moduleD (matched by glob pattern /abcz.moduleD/)" in longrepr
    assert "abcz.moduleA ↣ abcz.moduleB ↣ abcz.moduleC ↣ abcz.moduleD" in longrepr


def test_resolution_non_existing_module(create_testset):
    create_testset(
        ("abcz/__init__.py", ""),
        ("abcz/moduleA.py", "import abcz.moduleB"),
        ("abcz/moduleB.py", "from i_do_not_exist import i_also_do_not_exist\nimport abcz.moduleC"),
        ("abcz/moduleC.py", ""),
    )
    archrule("rule exclusion").match("abcz.moduleA").should_import("abcz.moduleC").check("abcz")



import importlib.resources
import sys
from functools import lru_cache
from typing import FrozenSet


@lru_cache
def core_modules(version=None) -> FrozenSet[str]:
    if version is None:
        version = sys.version_info
    if version >= (3, 10):
        modules = sys.stdlib_module_names | set(sys.builtin_module_names)  # type: ignore[attr-defined]
    else:
        modules_file = _module_file_path(version)
        if not modules_file.exists():
            raise FileNotFoundError(
                f"{modules_file} does not exist, perhaps your python version is not supported"
            )
        modules = frozenset(modules_file.read_text().splitlines())
    return modules


def _module_file_path(version):
    if sys.version_info < (3, 9):
        with importlib.resources.path("pytest_archon", "assets") as p:
            return p / "core-module-lists" / f"{version[0]}.{version[1]}.txt"
    return (
        importlib.resources.files("pytest_archon")
        / "assets"
        / "core-module-lists"
        / f"{version[0]}.{version[1]}.txt"
    )



from __future__ import annotations

import re
from dataclasses import dataclass
from fnmatch import fnmatchcase
from types import ModuleType
from typing import Callable, Set


from pytest_archon.collect import (
    ImportMap,
    collect_imports,
    recurse_imports,
    walk,
    walk_runtime,
    walk_toplevel,
)
from pytest_archon.failure import add_failure  # type: ignore[import]


ConstraintPredicate = Callable[[str, Set[str], ImportMap], bool]


@dataclass
class RulePattern:
    is_regex: bool
    pattern: str

    def match(self, k: str):
        if self.is_regex:
            return re.search(self.pattern, k)
        else:
            return fnmatchcase(k, self.pattern)

    def __str__(self):
        if self.is_regex:
            return f"regex pattern /{self.pattern}/"
        else:
            return f"glob pattern /{self.pattern}/"


def _as_rule_patterns(use_regex, patterns):
    return [RulePattern(is_regex=use_regex, pattern=p) for p in patterns]


def archrule(name: str, comment: str | None = None, *, use_regex: bool = False) -> Rule:
    """Define a new architectural rule with a name and an optional comment."""
    return Rule(name, comment=comment, use_regex=use_regex)


# https://peps.python.org/pep-0451/
# the path is the package path: where the submodules are in
class Rule:
    def __init__(self, name: str, comment: str | None, use_regex: bool = False) -> None:
        """Define a new architectural rule with a name and a comment."""
        self.name = name
        self.comment = comment
        self.use_regex = use_regex

    def match(self, *pattern: str, **kwargs) -> RuleTargets:
        """A glob pattern for modules this rule should match."""
        return RuleTargets(self).match(*pattern, **kwargs)

    def exclude(self, *pattern: str, **kwargs) -> RuleTargets:
        """A glob pattern for modules this rule should exclude from matching.

        Exclusion takes precedence of matching.
        """
        return RuleTargets(self).exclude(*pattern, **kwargs)


class RuleTargets:
    def __init__(self, rule: Rule) -> None:
        self.rule = rule
        self.match_criteria: list[RulePattern] = []
        self.exclude_criteria: list[RulePattern] = []

    def match(self, *pattern: str) -> RuleTargets:
        """A glob pattern for modules this rule should match."""

        self.match_criteria.extend(_as_rule_patterns(self.rule.use_regex, pattern))
        return self

    def exclude(self, *pattern: str) -> RuleTargets:
        """A glob pattern for modules this rule should exclude from matching.

        Exclusion takes precedence of matching.
        """
        self.exclude_criteria.extend(_as_rule_patterns(self.rule.use_regex, pattern))
        return self

    def should_not_import(self, *pattern: str, **kwargs) -> RuleConstraints:
        """Define a constraint that the defined modules should
        not import modules that match the given pattern.

        Keep in mind that module dependencies are checked transtively.

        E.g. 'mymodule.submodule', 'mymodule.*'
        """
        return RuleConstraints(self.rule, self).should_not_import(*pattern, **kwargs)

    def should(self, pred: ConstraintPredicate, name=None) -> RuleConstraints:
        """Define a constraint using a supplied function/predicate.

        If the predicate function returns ``False``, the constraint is
        considered as being violated.

        :param pred: a function with the signature
            ``pred(module, direct_imports, all_imports)``
        :type pred: ConstraintPredicate
        :param name: An optional name for this constraint. If omitted,
            the name will be inferred from the predicate function.
        :rtype: RuleConstraints
        """
        return RuleConstraints(self.rule, self).should(pred, name)

    def should_import(self, *pattern: str, **kwargs) -> RuleConstraints:
        """Define a constraint that the defined modules should
        import modules that match the given pattern.

        Keep in mind that module dependencies are checked transtively.

        E.g. 'mymodule.submodule', 'mymodule.*'
        """
        return RuleConstraints(self.rule, self).should_import(*pattern, **kwargs)

    def may_import(self, *pattern: str, **kwargs) -> RuleConstraints:
        """Loosen the constraints from should_import and
        should_not_import: modules matching may_import are
        excluded/ignored from the constraint check.
        """
        return RuleConstraints(self.rule, self).may_import(*pattern, **kwargs)


class RuleConstraints:
    def __init__(self, rule: Rule, targets: RuleTargets) -> None:
        self.rule = rule
        self.targets = targets
        self.forbidden: list[RulePattern] = []
        self.required: list[RulePattern] = []
        self.ignored: list[RulePattern] = []
        self.constraint_preds: list[tuple[ConstraintPredicate, str]] = []

    def should_not_import(self, *pattern: str) -> RuleConstraints:
        """Define a constraint that the defined modules should
        not import modules that match the given pattern.

        Keep in mind that module dependencies are checked transtively.

        E.g. 'mymodule.submodule', 'mymodule.*'
        """
        self.forbidden.extend(_as_rule_patterns(self.rule.use_regex, pattern))
        return self

    def should_import(self, *pattern: str) -> RuleConstraints:
        """Define a constraint that the defined modules should
        import modules that match the given pattern.

        Keep in mind that module dependencies are checked transtively.

        E.g. 'mymodule.submodule', 'mymodule.*'
        """
        self.required.extend(_as_rule_patterns(self.rule.use_regex, pattern))
        return self

    def should(self, pred: ConstraintPredicate, name=None):
        self.constraint_preds.append((pred, name or pred.__name__))
        return self

    def may_import(self, *pattern: str) -> RuleConstraints:
        """Loosen the constraints from should_import and
        should_not_import: modules matching may_import are
        excluded/ignored from the constraint check.
        """
        self.ignored.extend(_as_rule_patterns(self.rule.use_regex, pattern))
        return self

    def check(
        self,
        package: str | ModuleType,
        *,
        skip_type_checking=False,
        only_toplevel_imports=False,
        only_direct_imports=False,
    ) -> None:
        """Check the rule against a package or module.

        Options:

        skip_type_checking:
           Do not check TYPE_CHECKING blocks, used by static code analysers
        only_toplevel_imports:
            Do not traverse functions and methods, looking for imports
        only_direct_imports:
            Only check imports done by the module, not indirect imports
        """
        rule_name = self.rule.name
        rule_comment = self.rule.comment

        if only_toplevel_imports:
            walker = walk_toplevel
        elif skip_type_checking:
            walker = walk_runtime
        else:
            walker = walk

        all_imports = collect_imports(
            package,
            walker,
        )
        match_criteria = self.targets.match_criteria
        exclude_criteria = self.targets.exclude_criteria

        candidates: list[str] = []
        for mp in match_criteria:
            candidates.extend(k for k in all_imports.keys() if mp.match(k))
        for ep in exclude_criteria:
            candidates = [k for k in candidates if not ep.match(k)]

        match_criteria_pretty = [str(c) for c in match_criteria]
        exclude_criteria_pretty = [str(c) for c in exclude_criteria]
        if not candidates:
            add_failure(
                rule_name,
                rule_comment,
                f"NO CANDIDATES MATCHED. Match criteria: {match_criteria_pretty}, "
                f"exclude_criteria: {exclude_criteria_pretty}",
            )
            return

        candidates = sorted(candidates)

        for candidate in candidates:
            import_map = {candidate: all_imports[candidate]} if only_direct_imports else all_imports

            for constraint in self._find_required_constraints(candidate, import_map):
                add_failure(
                    rule_name,
                    rule_comment,
                    f"module '{candidate}' is missing REQUIRED imports matching {constraint}",
                    ["."],
                )

            for constraint, path in self._find_forbidden_constraints(candidate, import_map):
                add_failure(
                    rule_name,
                    rule_comment,
                    f"module '{candidate}' has FORBIDDEN import {path[-1]} (matched by {constraint})",
                    path,
                )

            for constraint in self._find_constraint_predicates(candidate, import_map):
                add_failure(
                    rule_name,
                    rule_comment,
                    f"module '{candidate}' VIOLATED constraint '{constraint}'",
                    ["."],
                )

    def _find_required_constraints(self, module: str, all_imports: ImportMap):
        for constraint in self.required:
            if not any(
                imp for path in recurse_imports(module, all_imports) if constraint.match(imp := path[-1])
            ):
                yield constraint

    def _find_constraint_predicates(self, module: str, all_imports: ImportMap):
        direct_imports = all_imports[module]
        for pred, name in self.constraint_preds:
            if not pred(module, direct_imports, all_imports):
                yield name

    def _find_forbidden_constraints(
        self,
        module: str,
        all_imports: ImportMap,
    ):
        for constraint in self.forbidden:
            yield from (
                (constraint, path)
                for path in recurse_imports(module, all_imports)
                if constraint.match(path[-1]) and not any(ignore.match(path[-1]) for ignore in self.ignored)
            )



from textwrap import dedent

import pytest

pytest_plugins = ["pytester"]


@pytest.fixture
def create_testset(tmp_path, monkeypatch):
    def _create_testset(*module_contents):
        for module, contents in module_contents:
            mod_path = tmp_path / module
            mod_path.parent.mkdir(parents=True, exist_ok=True)
            mod_path.write_text(dedent(contents))
        monkeypatch.syspath_prepend(str(tmp_path))
        return tmp_path

    return _create_testset



def test_rule_basic(archrule):
    (
        archrule("abc", "def")
        .match("*collect")
        .should_not_import("pytest_archon.failure")
        .check("pytest_archon")
    )


def test_rule_fail(pytester):
    pytester.makepyfile(
        """
        import pytest_archon

        def test_rule_fail(archrule):
            (
                archrule("abc", "def")
                .match("*collect")
                .should_not_import("importl*")
                .check(pytest_archon)
            )
    """
    )
    result = pytester.runpytest()
    result.assert_outcomes(failed=1)
    result.stdout.fnmatch_lines("FAILED Rule 'abc':")



"""
This code is based on pytest-check, by Bryan Okken (MIT License)
https://github.com/okken/pytest-check.
"""

from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class Failure:
    rule_name: str
    rule_comment: str
    reason: str
    path: tuple[str, ...]


_failures: list[Failure] = []


def add_failure(rule_name, rule_comment, reason, path: list[str] | None = None):
    global _failures
    _failures.append(Failure(rule_name, rule_comment, reason, tuple(path) if path else ()))


def pop_failures():
    global _failures
    try:
        return _failures
    finally:
        _failures = []



from pathlib import Path

from pytest_archon.collect import (
    collect_imports_from_path,
    path_to_module,
    recurse_imports,
    resolve_module_or_object_by_path,
    resolve_module_or_object_by_spec,
)


def test_collect_modules(create_testset):
    path = create_testset(("mymodule.py", ""))
    collected = list(name for name, _ in collect_imports_from_path(path, "pkg"))
    assert "pkg.mymodule" in collected


def test_collect_with_system_modules(create_testset):
    path = create_testset(("mymodule.py", "import sys, os"))

    name, imports = next(iter(collect_imports_from_path(path, "pkg")))

    assert name == "pkg.mymodule"
    assert "sys" in imports
    assert "os" in imports


def test_path_to_module():
    assert path_to_module(Path("a/b/./c/d/../e"), Path("a/b/c")) == "d.e"


def test_module_imports_other_module(create_testset):
    path = create_testset(("module.py", ""), ("othermodule.py", "import module"))

    module_names = {i for name, imports in collect_imports_from_path(path, "pkg") for i in imports}

    assert "module" in module_names
    assert "othermodule" not in module_names


def test_module_import_from(create_testset):
    path = create_testset(("module.py", "val = 1"), ("othermodule.py", "from module import val"))

    module_names = {i for name, imports in collect_imports_from_path(path, "pkg") for i in imports}
    assert module_names == {"module"}


def test_module_import_nested_modules(create_testset):
    path = create_testset(
        ("package/__init__.py", ""),
        ("package/module.py", ""),
        ("package/othermodule.py", "import package.module"),
    )

    module_names = {i for name, imports in collect_imports_from_path(path, "package") for i in imports}

    assert "package.module" in module_names
    assert "package.othermodule" not in module_names
    assert "othermodule" not in module_names


def test_relative_imports(create_testset):
    path = create_testset(
        ("package/__init__.py", ""),
        ("package/module.py", "from .importme import val"),
        ("package/importme.py", "val = 1"),
        ("package/sub/__init__.py", ""),
        ("package/sub/deep.py", "from ..importme import val"),
    )

    module_names = {
        i for name, imports in collect_imports_from_path(path / "package", "package") for i in imports
    }

    assert "package.importme" in module_names


def test_namespace_pkgs(create_testset):
    create_testset(
        ("package/__init__.py", ""),
        ("package/initless/module.py", "A=3"),
    )

    res = resolve_module_or_object_by_path("package.initless.module.A")
    assert res == "package.initless.module"


def test_resolve_module_or_object_by_spec():
    res = resolve_module_or_object_by_spec("fnmatch.fnmatch")
    assert res == "fnmatch"


def test_resolve_module_or_object_by_path():
    res = resolve_module_or_object_by_path("fnmatch.fnmatch")
    assert res == "fnmatch"


def test_recurse_imports():
    all_imports = {"a": ["b", "c"], "b": ["c", "d"], "c": ["e"]}
    res = list(recurse_imports("a", all_imports))

    assert res == [("a", "b"), ("a", "b", "c"), ("a", "b", "c", "e"), ("a", "b", "d"), ("a", "c")]



import pytest

from pytest_archon.rule import Rule, RuleConstraints, RuleTargets


@pytest.mark.parametrize(
    "method1,method2",
    [
        (Rule.match, RuleTargets.match),
        (Rule.exclude, RuleTargets.exclude),
        (RuleTargets.should_import, RuleConstraints.should_import),
        (RuleTargets.should_not_import, RuleConstraints.should_not_import),
    ],
)
def test_docstrings_match(method1, method2):
    assert method1.__doc__ == method2.__doc__



from __future__ import annotations

import ast
import os
import re
import sys
from collections import deque
from functools import lru_cache
from importlib.util import find_spec
from logging import getLogger
from pathlib import Path
from types import ModuleType
from typing import Callable, Dict, Iterable, Iterator, Sequence, Set

from pytest_archon.core_modules import core_modules

# https://docs.djangoproject.com/en/4.1/_modules/django/utils/module_loading/
# https://stackoverflow.com/questions/54325116/can-i-handle-imports-in-an-abstract-syntax-tree
# https://bugs.python.org/issue38721
# https://github.com/0cjs/sedoc/blob/master/lang/python/importers.md


Walker = Callable[[ast.Module], Iterator[ast.AST]]
ImportMap = Dict[str, Set[str]]

logger = getLogger(__name__)


def collect_imports(package: str | ModuleType, walker: Walker) -> ImportMap:
    if isinstance(package, ModuleType):
        if not hasattr(package, "__path__"):
            raise AttributeError(f"module {package.__name__} does not have __path__")
        package = package.__name__

    all_imports: ImportMap = {}
    for name, imports in collect_imports_from_path(package_dir(package), package, walker):
        direct_imports = {imp for imp in imports if imp != name}
        if name in all_imports:
            raise KeyError(f"WTF? duplicate module {name}")
        all_imports[name] = direct_imports
    return all_imports


def walk(node: ast.AST) -> Iterator[ast.AST]:
    return ast.walk(node)


def walk_runtime(node: ast.AST) -> Iterator[ast.AST]:
    """Skip TYPE_CHECKING markers.

    The check if pretty rudimentary:
    it checks for if statements with either TYPE_CHECKING or
    <somemod>.TYPECHECKING in the expression.
    """
    todo = deque([node])
    while todo:
        node = todo.popleft()
        if not type_checking_clause(node):
            todo.extend(ast.iter_child_nodes(node))
            yield node


def walk_toplevel(node: ast.Module) -> Iterator[ast.AST]:
    yield from node.body


def package_dir(package: str) -> Path:
    spec = find_spec(package)
    if not spec:
        raise ModuleNotFoundError(f"could not find the module {package!r}", name=package)

    assert spec.origin
    return Path(spec.origin).parent


@lru_cache(maxsize=2048)
def collect_imports_from_path(
    path: Path, package: str, walker: Walker = walk
) -> frozenset[tuple[str, frozenset[str]]]:
    def _collect(py_file):
        module_name = path_to_module(py_file, path, package)
        tree = ast.parse(py_file.read_bytes())
        imports = extract_imports_ast(walker(tree), module_name)
        return module_name, frozenset(imports)

    return frozenset(_collect(py_file) for py_file in Path(path).glob("**/*.py"))


def path_to_module(module_path: Path, base_path: Path, package=None) -> str:
    rel_path = module_path.relative_to(base_path)

    if package:
        parts = [package]
    else:
        parts = []
    parts.extend(rel_path.parent.parts)

    if rel_path.stem != "__init__":
        parts.append(rel_path.stem)
    module = ".".join(parts)
    # some very basic sanitation
    return re.sub(r"\.+", ".", module)


def extract_imports_ast(nodes: Iterator[ast.AST], package: str, resolve=True) -> Iterator[str]:
    for node in nodes:
        if isinstance(node, ast.Import):
            for alias in node.names:
                yield alias.name
        elif isinstance(node, ast.ImportFrom):
            for alias in node.names:
                fqname = resolve_import_from(alias.name, node.module, package=package, level=node.level)
                if resolve:
                    try:
                        yield resolve_module_or_object_by_path(fqname)
                    except ModuleNotFoundError:
                        logger.warning(
                            f"Could not determine if {fqname} is a module or function."
                            " Assuming it is a module."
                        )
                        yield fqname
                else:
                    yield fqname


def type_checking_clause(node: ast.AST) -> bool:
    return isinstance(node, ast.If) and (
        (isinstance(node.test, ast.Name) and node.test.id == "TYPE_CHECKING")
        or (isinstance(node.test, ast.Attribute) and node.test.attr == "TYPE_CHECKING")
    )


# TODO replace with importlib.util.resolve_name ?
def resolve_import_from(
    name: str, module: str | None = None, package: str | None = None, level: int = 0
) -> str:
    if not level:
        # absolute import
        if name == "*":
            assert module
            return module
        return name if module is None else f"{module}.{name}"

    # taken from importlib._bootstrap._resolve_name
    assert package
    bits = package.rsplit(".", level)
    if len(bits) < level:
        raise ImportError("attempted relative import beyond top-level package")
    base = bits[0]

    # relative import
    if module is None:
        # from . import moduleX
        prefix = base
    else:
        # from .moduleZ import moduleX
        prefix = f"{base}.{module}"
    if name == "*":
        return prefix
    return f"{prefix}.{name}"


# I'm not sure what to do with this function.
# importlib.util.find_spec is flaky and sometimes doesn't work
# for now we use resolve_module_or_object_by_path till I figure
# out what the issue is
def resolve_module_or_object_by_spec(fqname: str) -> str:
    if fqname in core_modules():
        return fqname

    parent_name = fqname.rpartition(".")[0]

    # shortcut to deal with e.g. from __future__ import annotations
    if parent_name in core_modules():
        return parent_name

    spec = None
    try:
        spec = find_spec(fqname)
    except ModuleNotFoundError as ex:
        parent_spec = find_spec(parent_name)
        # if we cannot find the parent, then something is off
        if not parent_spec:
            raise ex

    return fqname if spec else fqname.rpartition(".")[0]


def resolve_module_or_object_by_path(fqname: str) -> str:
    if "." not in fqname:
        return fqname

    if fqname in core_modules():
        return fqname

    parts = fqname.split(".")
    head = parts[0]
    parent_name = ".".join(parts[:-1])

    # shortcut to deal with e.g. from __future__ import annotations
    if parent_name in core_modules():
        return parent_name

    # taken from importlib.util.find_spec
    if fqname in sys.modules:
        module = sys.modules[fqname]
        if module is None:
            return None
        try:
            spec = module.__spec__
        except AttributeError:
            raise ValueError(f"{fqname}.__spec__ is not set") from None
        else:
            if spec is None:
                raise ValueError(f"{fqname}.__spec__ is None")
            return fqname

    spec = find_spec(head)
    if not spec:
        raise ModuleNotFoundError(f"could not find the module {head} to resolve {fqname}", name=head)

    has_no_submodules = (
        not hasattr(spec, "submodule_search_locations") or spec.submodule_search_locations is None
    )
    if len(parts) == 2 and has_no_submodules:
        # we have from a import b with a being a.py, is that possible:
        # a package without directories, just one file? I assume yes for now
        return head

    if has_no_submodules:
        raise ModuleNotFoundError(f"could not find the module {head} to resolve {fqname}", name=head)

    assert spec.submodule_search_locations
    for base_path in spec.submodule_search_locations:
        init = os.path.join(base_path, *(parts[1:] + ["__init__.py"]))
        dir = os.path.join(base_path, *parts[1:])
        direct = os.path.join(base_path, *parts[1:]) + ".py"
        if os.path.exists(init) or os.path.exists(direct) or os.path.isdir(dir):
            # we have a module
            return fqname
    # we imported an object, return "parent"
    return parent_name


def recurse_imports(module: str, all_imports: ImportMap) -> Iterable[Sequence[str]]:
    seen = set()

    def recurse(path):
        mod = path[-1]
        if mod in seen or mod not in all_imports:
            return

        seen.add(mod)
        for imp in all_imports[mod]:
            new_path = path + (imp,)
            yield new_path
            yield from recurse(new_path)

    yield from recurse((module,))



from itertools import groupby
from operator import attrgetter

import pytest
from _pytest._code.code import ExceptionInfo

from pytest_archon.failure import pop_failures
from pytest_archon.rule import archrule


@pytest.fixture(name="archrule")
def check_fixture():
    return archrule


from _pytest.skipping import xfailed_key


@pytest.hookimpl(hookwrapper=True, trylast=True)
def pytest_runtest_makereport(item, call):
    outcome = yield
    report = outcome.get_result()

    failures = pop_failures()

    if failures and item._store[xfailed_key]:
        report.outcome = "skipped"
        report.wasxfail = item._store[xfailed_key].reason
        return

    longrepr = format_failures(failures)
    if longrepr:
        report.longrepr = longrepr
        report.outcome = "failed"

    if failures:
        try:
            raise AssertionError(report.longrepr)
        except AssertionError:
            excinfo = ExceptionInfo.from_current()
        call.excinfo = excinfo


def format_failures(failures):
    longrepr = []
    for rule_name, rule_failures in groupby(failures, attrgetter("rule_name")):
        longrepr.append(f"FAILED Rule '{rule_name}':")
        for reason, reason_failures in groupby(rule_failures, attrgetter("reason")):
            longrepr.append(f"- {reason}")
            for path in set(f.path for f in reason_failures):
                longrepr.append(f"    from {' ↣ '.join(path)}")
    return "\n".join(longrepr)


class ModelViolation(AssertionError):
    pass



import ast
from textwrap import dedent

from pytest_archon.collect import extract_imports_ast, walk, walk_runtime, walk_toplevel


def test_parse_imports():
    code = dedent(
        """\
        import sys
        if True:
            from datetime import datetime
        """
    )

    root = ast.parse(code, "test_parse.py")
    imports = list(extract_imports_ast(walk(root), ""))

    assert "sys" in imports
    assert "datetime" in imports


def test_parse_toplevel_imports():
    code = dedent(
        """\
        import sys
        if 0:
            from datetime import datetime
        """
    )

    root = ast.parse(code, "test_parse.py")
    imports = list(
        extract_imports_ast(
            walk_toplevel(root),
            "",
        )
    )

    assert "sys" in imports
    assert "datetime" not in imports


def test_skip_type_checking_marker():
    code = dedent(
        """\
        import sys
        from typing import TYPE_CHECKING
        if TYPE_CHECKING:
            from datetime import datetime
        """
    )

    root = ast.parse(code, "test_parse.py")
    imports = list(extract_imports_ast(walk_runtime(root), ""))

    # Should this be os.path?
    assert "datetime" not in imports
    assert "sys" in imports


def test_skip_typing_dot_type_checking_marker():
    code = dedent(
        """\
        import sys
        import typing
        if typing.TYPE_CHECKING:
            from datetime import datetime
        """
    )

    root = ast.parse(code, "test_parse.py")
    imports = list(extract_imports_ast(walk_runtime(root), ""))

    # Should this be os.path?
    assert "datetime" not in imports
    assert "sys" in imports


def test_parse_relative_imports(create_testset, monkeypatch):
    path = create_testset(
        ("pkgA/subpkg2/moduleZ.py", ""),
        ("pkgA/subpkg2/__init__.py", ""),
        ("pkgA/subpkg2/subpkg2a/moduleM.py", ""),
        ("pkgA/subpkg2/subpkg2a/__init__.py", ""),
        ("pkgA/__init__.py", ""),
        ("pkgA/moduleA.py", ""),
        ("pkgA/subpkg1/__init__.py", ""),
        ("pkgA/subpkg1/moduleY.py", ""),
        ("pkgA/subpkg1/subpkg1a/moduleL.py", ""),
        ("pkgA/subpkg1/subpkg1a/__init__.py", ""),
        ("pkgA/subpkg1/subpkg1a/moduleK.py", ""),
        ("pkgA/subpkg1/moduleX.py", ""),
    )

    code = dedent(
        """\
        import datetime
        import pkgA
        import pkgA.subpkg1
        import pkgA.subpkg1.subpkg1a
        from pkgA import subpkg2
        from . import variableA, variableB
        from .. import moduleA
        from . import moduleX, moduleY
        from ..subpkg2.subpkg2a import moduleM
        from ..subpkg2.moduleZ import CONSTANT_A
        """
    )

    monkeypatch.syspath_prepend(path)
    root = ast.parse(code)
    imports = set(extract_imports_ast(walk(root), "pkgA.subpkg1.subpkg1a"))

    assert imports == {
        "datetime",
        "pkgA",
        "pkgA.subpkg1",
        "pkgA.subpkg1.subpkg1a",
        "pkgA.subpkg2",
        "pkgA.moduleA",
        "pkgA.subpkg1.moduleX",
        "pkgA.subpkg1.moduleY",
        "pkgA.subpkg2.subpkg2a.moduleM",
        "pkgA.subpkg2.moduleZ",
    }



import sys

import pytest

import pytest_archon.core_modules as cm


def test_read_legacy():
    modules = cm.core_modules((3, 9))
    assert "__future__" in modules


@pytest.mark.skipif(sys.version_info < (3, 10), reason="requires python3.10 or higher")
def test_read():
    modules = cm.core_modules((3, 10))
    assert "__future__" in modules


def test_read_error():
    with pytest.raises(FileNotFoundError):
        cm.core_modules((3, 5))
